/*
default theme
https://github.com/richleland/pygments-css/blob/master/default.css
*/

body[data-md-color-scheme="default"] div.highlight code { border-left: medium solid green}

body[data-md-color-scheme="default"] .highlight .hll { background-color: #ffffcc }
body[data-md-color-scheme="default"] .highlight  { background: #f8f8f8; }
body[data-md-color-scheme="default"] .highlight .c { color: #408080; font-style: italic } /* Comment */
body[data-md-color-scheme="default"] .highlight .err { border: 1px solid #FF0000 } /* Error */
body[data-md-color-scheme="default"] .highlight .k { color: #008000; font-weight: bold } /* Keyword */
body[data-md-color-scheme="default"] .highlight .o { color: #666666 } /* Operator */
body[data-md-color-scheme="default"] .highlight .ch { color: #408080; font-style: italic } /* Comment.Hashbang */
body[data-md-color-scheme="default"] .highlight .cm { color: #408080; font-style: italic } /* Comment.Multiline */
body[data-md-color-scheme="default"] .highlight .cp { color: #BC7A00 } /* Comment.Preproc */
body[data-md-color-scheme="default"] .highlight .cpf { color: #408080; font-style: italic } /* Comment.PreprocFile */
body[data-md-color-scheme="default"] .highlight .c1 { color: #408080; font-style: italic } /* Comment.Single */
body[data-md-color-scheme="default"] .highlight .cs { color: #408080; font-style: italic } /* Comment.Special */
body[data-md-color-scheme="default"] .highlight .gd { color: #A00000 } /* Generic.Deleted */
body[data-md-color-scheme="default"] .highlight .ge { font-style: italic } /* Generic.Emph */
body[data-md-color-scheme="default"] .highlight .gr { color: #FF0000 } /* Generic.Error */
body[data-md-color-scheme="default"] .highlight .gh { color: #000080; font-weight: bold } /* Generic.Heading */
body[data-md-color-scheme="default"] .highlight .gi { color: #00A000 } /* Generic.Inserted */
body[data-md-color-scheme="default"] .highlight .go { color: #888888 } /* Generic.Output */
body[data-md-color-scheme="default"] .highlight .gp { color: #000080; font-weight: bold } /* Generic.Prompt */
body[data-md-color-scheme="default"] .highlight .gs { font-weight: bold } /* Generic.Strong */
body[data-md-color-scheme="default"] .highlight .gu { color: #800080; font-weight: bold } /* Generic.Subheading */
body[data-md-color-scheme="default"] .highlight .gt { color: #0044DD } /* Generic.Traceback */
body[data-md-color-scheme="default"] .highlight .kc { color: #008000; font-weight: bold } /* Keyword.Constant */
body[data-md-color-scheme="default"] .highlight .kd { color: #008000; font-weight: bold } /* Keyword.Declaration */
body[data-md-color-scheme="default"] .highlight .kn { color: #008000; font-weight: bold } /* Keyword.Namespace */
body[data-md-color-scheme="default"] .highlight .kp { color: #008000 } /* Keyword.Pseudo */
body[data-md-color-scheme="default"] .highlight .kr { color: #008000; font-weight: bold } /* Keyword.Reserved */
body[data-md-color-scheme="default"] .highlight .kt { color: #B00040 } /* Keyword.Type */
body[data-md-color-scheme="default"] .highlight .m { color: #666666 } /* Literal.Number */
body[data-md-color-scheme="default"] .highlight .s { color: #BA2121 } /* Literal.String */
body[data-md-color-scheme="default"] .highlight .na { color: #7D9029 } /* Name.Attribute */
body[data-md-color-scheme="default"] .highlight .nb { color: #008000 } /* Name.Builtin */
body[data-md-color-scheme="default"] .highlight .nc { color: #0000FF; font-weight: bold } /* Name.Class */
body[data-md-color-scheme="default"] .highlight .no { color: #880000 } /* Name.Constant */
body[data-md-color-scheme="default"] .highlight .nd { color: #AA22FF } /* Name.Decorator */
body[data-md-color-scheme="default"] .highlight .ni { color: #999999; font-weight: bold } /* Name.Entity */
body[data-md-color-scheme="default"] .highlight .ne { color: #D2413A; font-weight: bold } /* Name.Exception */
body[data-md-color-scheme="default"] .highlight .nf { color: #0000FF } /* Name.Function */
body[data-md-color-scheme="default"] .highlight .nl { color: #A0A000 } /* Name.Label */
body[data-md-color-scheme="default"] .highlight .nn { color: #0000FF; font-weight: bold } /* Name.Namespace */
body[data-md-color-scheme="default"] .highlight .nt { color: #008000; font-weight: bold } /* Name.Tag */
body[data-md-color-scheme="default"] .highlight .nv { color: #19177C } /* Name.Variable */
body[data-md-color-scheme="default"] .highlight .ow { color: #AA22FF; font-weight: bold } /* Operator.Word */
body[data-md-color-scheme="default"] .highlight .w { color: #bbbbbb } /* Text.Whitespace */
body[data-md-color-scheme="default"] .highlight .mb { color: #666666 } /* Literal.Number.Bin */
body[data-md-color-scheme="default"] .highlight .mf { color: #666666 } /* Literal.Number.Float */
body[data-md-color-scheme="default"] .highlight .mh { color: #666666 } /* Literal.Number.Hex */
body[data-md-color-scheme="default"] .highlight .mi { color: #666666 } /* Literal.Number.Integer */
body[data-md-color-scheme="default"] .highlight .mo { color: #666666 } /* Literal.Number.Oct */
body[data-md-color-scheme="default"] .highlight .sa { color: #BA2121 } /* Literal.String.Affix */
body[data-md-color-scheme="default"] .highlight .sb { color: #BA2121 } /* Literal.String.Backtick */
body[data-md-color-scheme="default"] .highlight .sc { color: #BA2121 } /* Literal.String.Char */
body[data-md-color-scheme="default"] .highlight .dl { color: #BA2121 } /* Literal.String.Delimiter */
body[data-md-color-scheme="default"] .highlight .sd { color: #BA2121; font-style: italic } /* Literal.String.Doc */
body[data-md-color-scheme="default"] .highlight .s2 { color: #BA2121 } /* Literal.String.Double */
body[data-md-color-scheme="default"] .highlight .se { color: #BB6622; font-weight: bold } /* Literal.String.Escape */
body[data-md-color-scheme="default"] .highlight .sh { color: #BA2121 } /* Literal.String.Heredoc */
body[data-md-color-scheme="default"] .highlight .si { color: #BB6688; font-weight: bold } /* Literal.String.Interpol */
body[data-md-color-scheme="default"] .highlight .sx { color: #008000 } /* Literal.String.Other */
body[data-md-color-scheme="default"] .highlight .sr { color: #BB6688 } /* Literal.String.Regex */
body[data-md-color-scheme="default"] .highlight .s1 { color: #BA2121 } /* Literal.String.Single */
body[data-md-color-scheme="default"] .highlight .ss { color: #19177C } /* Literal.String.Symbol */
body[data-md-color-scheme="default"] .highlight .bp { color: #008000 } /* Name.Builtin.Pseudo */
body[data-md-color-scheme="default"] .highlight .fm { color: #0000FF } /* Name.Function.Magic */
body[data-md-color-scheme="default"] .highlight .vc { color: #19177C } /* Name.Variable.Class */
body[data-md-color-scheme="default"] .highlight .vg { color: #19177C } /* Name.Variable.Global */
body[data-md-color-scheme="default"] .highlight .vi { color: #19177C } /* Name.Variable.Instance */
body[data-md-color-scheme="default"] .highlight .vm { color: #19177C } /* Name.Variable.Magic */
body[data-md-color-scheme="default"] .highlight .il { color: #666666 } /* Literal.Number.Integer.Long */
